from __future__ import annotations

import logging
import re
from typing import Any

import pygls.lsp.types as lsp_types
from pygls.workspace import Document

from yls import completion
from yls import utils
from yls.completion import CONDITION_KEYWORDS
from yls.plugin_manager_provider import PluginManagerProvider
from yls.strings import estimate_string_type
from yls.strings import string_modifiers_completion_items

log = logging.getLogger(__name__)


class Completer:

    REGEXP_META = re.compile(r'^\s*(\w+)\s+=\s+"(.*)$')
    REGEXP_IMPORT = re.compile(r'^import\s+"(.*)$')

    def __init__(self, ls: Any):
        self.ls = ls
        self.completion_cache = completion.CompletionCache.from_yaramod(self.ls.ymod)

    def complete(self, params: lsp_types.CompletionParams) -> lsp_types.CompletionList:
        return lsp_types.CompletionList(is_incomplete=False, items=self._complete(params))

    def signature_help(self, params: lsp_types.CompletionParams) -> lsp_types.SignatureHelp | None:
        signatures = self._signature_help(params)
        if signatures is None:
            return None

        return lsp_types.SignatureHelp(
            signatures=signatures, active_parameter=0, active_signature=0
        )

    def _signature_help(
        self, params: lsp_types.CompletionParams
    ) -> list[lsp_types.SignatureInformation] | None:
        text_doc = self.ls.workspace.get_document(params.text_document.uri)
        word = utils.cursor_word(text_doc, params.position)
        log.debug(f'[SIGNATURE_HELP] Cursor word is "{word}"')
        if word is None:
            log.debug("[SIGNATURE_HELP] Aborting signature help request")
            return None

        # Try to extract the function call from the word
        word = re.sub(r"\(.*", "", word)
        word = word.replace("(", "").replace(")", "")
        log.debug(f'[SIGNATURE_HELP] Normalized cursor word is "{word}"')

        if not word:
            log.debug(
                "[SIGNATURE_HELP] Aborting signature help request - invalid word after normalization"
            )
            return None

        symbol = self.completion_cache.get_function(word)
        if not symbol:
            log.debug("[SIGNATURE_HELP] Aborting signature help request - no symbol found")
            return None

        info: list[lsp_types.SignatureInformation] = []

        # NOTE: Handler other overloads
        info.extend(symbol.to_signature_information(word))

        return info

    def _complete(self, params: lsp_types.CompletionParams) -> list[lsp_types.CompletionItem]:
        document = self.ls.workspace.get_document(params.text_document.uri)

        res = []

        # If the cursor is not on a word, we still want to provide completions
        # for root module
        word = utils.cursor_word(document, params.position) or ""

        # Import completion
        log.debug("[COMPLETION] Adding import completion")
        res += self.complete_import(document, params.position)

        # String modifiers completion
        log.debug("[COMPLETION] Adding string modifiers completion")
        res += self.complete_string_modifiers(document, params.position, word)

        # Condition keywords completion
        log.debug("[COMPLETION] Adding condition keywords completion")
        res += self.complete_condition_keywords(document, params.position, word)

        # Function completion
        log.debug("[COMPLETION] Adding module completion")
        res += self.complete_word(document, params.position, word)

        # Symbols from last yara file completion
        log.debug("[COMPLETION] Adding last valid yara file")
        res += self.complete_last_valid_yara_file(document, params, word)

        # Plugin completion
        log.debug("COMPLETION] Adding completion items from plugings")
        res += utils.flatten_list(
            PluginManagerProvider.instance().hook.yls_completion(params=params, document=document)
        )

        return res

    def complete_last_valid_yara_file(
        self, document: Document, params: lsp_types.CompletionParams, word: str
    ) -> list[lsp_types.CompletionItem]:
        """Return list of completion items from last valid YaraFile."""
        if self.ls.last_valid_yara_file is None:
            return []

        if not utils.is_in_yara_section(document, params.position.line, "condition"):
            return []

        res = []

        # Add rules
        log.debug("[COMPLETION] Adding symbols from last valid YARA file")
        res += [
            lsp_types.CompletionItem(label=rule.name, insert_text=rule.name, sort_text="zb")
            for rule in self.ls.last_valid_yara_file.rules
            if rule.name.startswith(word)
        ]

        rule = self.ls.get_current_rule(
            params.text_document.uri, params.position, self.ls.last_valid_yara_file
        )
        if rule is None:
            return res

        # Add current rule strings
        res += [
            lsp_types.CompletionItem(
                label=string.identifier, insert_text=string.identifier, sort_text="za"
            )
            for string in rule.strings
            if string.identifier.startswith(word)
        ]

        return res

    def complete_word(
        self, document: Document, position: lsp_types.Position, word: str
    ) -> list[lsp_types.CompletionItem]:
        """Complete a function."""
        log.debug(f'[COMPLETION] Cursor word is "{word}"')

        if not utils.is_in_yara_section(document, position.line, "condition"):
            return []

        res = []

        symbols = self.completion_cache.get_symbols_matching(word)

        for symbol in symbols:
            res.extend(symbol.to_completion_items())
            # symbol_doc = symbol_doc.strip().replace("\t", "")
            # symbol_doc = re.sub(r" {2,}", "", symbol_doc)

        return res

    def complete_import(
        self, document: Document, position: lsp_types.Position
    ) -> list[lsp_types.CompletionItem]:
        """Complete an import statement."""
        res = []

        line = utils.cursor_line(document, position)
        match_res = self.REGEXP_IMPORT.search(line)

        if match_res:
            log.debug("[COMPLETION] Completing import values")
            for module in self.completion_cache.modules.attributes.keys():
                res.append(
                    lsp_types.CompletionItem(
                        label=module,
                        insert_text=module,
                        sort_text="aaa",
                        kind=lsp_types.CompletionItemKind.Module,
                    )
                )

        return res

    def complete_string_modifiers(
        self, document: Document, position: lsp_types.Position, word: str
    ) -> list[lsp_types.CompletionItem]:
        """Complete string modifiers."""

        # Check if the line is not empty
        line = utils.cursor_line(document, position)
        if not line.strip():
            return []

        # Complete only if end of the string is before the cursor
        characters_after_cursor = set(line[position.character :])
        if {'"', "/", "}"}.intersection(characters_after_cursor):
            return []

        # Complete only in `strings:` section
        if not utils.is_in_yara_section(document, position.line, "strings"):
            return []

        # Filter out the modifiers based on the string type
        estimated_string_type = estimate_string_type(line)
        final_list = iter(string_modifiers_completion_items(estimated_string_type))

        # Filter the modifiers based on the word under the cursor
        final_list = filter(lambda item: item.label.startswith(word), final_list)
        return list(final_list)

    def complete_condition_keywords(
        self, document: Document, position: lsp_types.Position, word: str
    ) -> list[lsp_types.CompletionItem]:

        # Complete only in `condition:` section
        if not utils.is_in_yara_section(document, position.line, "condition"):
            return []

        res = []
        for keyword in CONDITION_KEYWORDS:
            if not keyword.startswith(word):
                continue

            item = lsp_types.CompletionItem(
                label=keyword,
                kind=lsp_types.CompletionItemKind.Keyword,
                insert_text=keyword,
                insert_text_format=lsp_types.InsertTextFormat.PlainText,
                sort_text="a",
            )

            res.append(item)
        return res



from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import Any

import pygls.lsp.types as lsp_types
from pygls.workspace import Document

from yls import utils
from yls.plugin_manager_provider import PluginManagerProvider

log = logging.getLogger(__name__)


def from_params(
    ls: Any, params: lsp_types.CodeActionParams
) -> list[lsp_types.Command | lsp_types.CodeAction] | None:
    """Convert the request params to list of `CodeAction`s."""
    log.debug(f"[CODE_ACTION] {params=}")

    doc = ls.workspace.get_document(params.text_document.uri)
    res: list[lsp_types.Command | lsp_types.CodeAction] = []

    # Try to create string extraction code actions
    res.extend(string_extraction(doc, params))

    # Get the code actions from plugins
    plugin_code_actions = utils.flatten_list(
        PluginManagerProvider.instance().hook.yls_code_actions(ls=ls, params=params)
    )
    res.extend(plugin_code_actions)

    return res


def string_extraction(
    doc: Document, params: lsp_types.CodeActionParams
) -> list[lsp_types.CodeAction]:
    res = []
    cursor_line_index = params.range.start.line
    cursor_line = utils.document_line(doc, cursor_line_index)

    if is_line_for_string_extraction(cursor_line) and utils.is_in_yara_section(
        doc, cursor_line_index, "strings"
    ):
        string_extraction = lines_for_string_extraction(doc, cursor_line_index)

        if not string_extraction.strings:
            return []

        first_new_string_index = estimate_first_new_string_index(doc, cursor_line_index)

        # Construct the new code
        new_text = ""
        for idx, line in enumerate(string_extraction.strings):
            line = line.strip()
            if not line.startswith('"'):
                line = f'"{line}'
            if not line.endswith('"'):
                line = f'{line}"'
            new_text += f"\t\t$s{first_new_string_index + idx:02} = {line}\n"

        # Strip the last newline from the string, for some reason it is
        # inserted automatically
        new_text = new_text.rstrip()

        edit = lsp_types.TextEdit(
            range=utils.range_from_lines(
                string_extraction.first_string_line,
                string_extraction.last_string_line,
                doc,
                strip_whitespace=False,
            ),
            new_text=new_text,
        )
        res.append(
            lsp_types.CodeAction(
                title="Extract as strings",
                edit=lsp_types.WorkspaceEdit(changes={doc.uri: [edit]}),
                kind=lsp_types.CodeActionKind.RefactorRewrite,
            )
        )

    return res


def is_line_for_string_extraction(line: str) -> bool:
    """Check if the string can be extracted from line."""
    return bool(re.match(r'((\s*".*"\s*)|(^(?!\s+((//)|(\$\w+ =)|([a-z]+:)|($)))))', line))


@dataclass
class StringExtraction:
    strings: list[str]
    first_string_line: int

    @property
    def last_string_line(self) -> int:
        return self.first_string_line + len(self.strings) - 1


def lines_for_string_extraction(doc: Document, cursor_line: int) -> StringExtraction:
    """Return a list of lines around `cursor_line` that contains extractable strings."""
    res: list[str] = []
    first_string_line = 0

    # Walk back from cursor position (cursor_line included)
    for i in reversed(range(0, cursor_line + 1)):
        if is_line_for_string_extraction(doc.lines[i]):
            res.insert(0, doc.lines[i])
            first_string_line = i
        else:
            break

    # Walk forward from cursor position
    for i in range(cursor_line + 1, len(doc.lines)):
        if is_line_for_string_extraction(doc.lines[i]):
            res.append(doc.lines[i])
        else:
            break

    return StringExtraction(res, first_string_line)


def estimate_first_new_string_index(doc: Document, cursor_line: int) -> int:
    """Estimate the index of the upcoming string in current string section of
    YARA rule.

    If we are not able to find a proper value return 0 as a neutral element.
    """

    # Walk back from cursor position (cursor_line included)
    for i in reversed(range(0, cursor_line + 1)):

        # If we encounter end of string section, stop searching
        if re.match(r"\s*strings:\s*", doc.lines[i]):
            break

        # Check if the line looks like a string id
        match = re.match(r'\s*\$s(?P<idx>\d+) = ".*"\s*', doc.lines[i], re.ASCII)
        if match:
            last_string_index = int(match.group("idx"))
            return last_string_index + 1

    # If we did not find a suitable `id`, return neutral element
    return 0



# pylint: disable=unused-argument

from __future__ import annotations

import logging

import yaramod

from yls.plugin_manager_provider import PluginManagerProvider

log = logging.getLogger(__name__)


class YaramodProvider:
    """Singleton class providing Yaramod parser object."""

    ymod = None

    @classmethod
    def instance(cls) -> yaramod.Yaramod:
        """Return singleton instance."""
        if cls.ymod is None:
            cls.ymod = PluginManagerProvider.instance().hook.create_yaramod_instance()
        return cls.ymod



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Any

import yari

from yls import utils
from yls.hookspecs import ErrorMessage
from yls.hookspecs import InfoMessage
from yls.hookspecs import PopupMessage

log = logging.getLogger(__name__)


class Debugger:
    def __init__(self) -> None:
        self.context_rule_name: str | None = None
        self.ctxs: dict[str, yari.Context] = {}

    async def set_context(self, ls: Any, sample_hash: str, ruleset: str) -> PopupMessage:
        log.debug(f"[DEBUGGER] Setting context on bundled YARI to {sample_hash}")
        self.ctxs = {}

        self.context_rule_name = re.findall(r"(?<=^rule )\w+(?= {$)", ruleset, re.MULTILINE)[-1]

        samples_dir = await self.get_samples_dir(ls)
        if samples_dir is None:
            return ErrorMessage("Samples folder is not set")

        files = list(samples_dir.rglob(f"{sample_hash}*"))

        log.debug(f'[DEBUGGER] Found sample and module files for hash "{sample_hash}": {files}')
        if not files:
            return ErrorMessage(f"Sample {sample_hash} not found in {samples_dir}")

        # Sort files into groups by type of source
        sample: str | None = None
        module: str | None = None
        for f in files:
            if f.name == sample_hash and not sample:
                sample = str(f)
            elif not module:
                module = str(f)
            else:
                break

        if sample:
            self.ctxs["SAMPLE"] = yari.Context(  # pylint: disable=no-member
                sample=sample, rule_string=ruleset
            )

            if module:
                self.ctxs["CUCKOO"] = yari.Context(  # pylint: disable=no-member
                    sample=sample, module_data={"cuckoo": module}, rule_string=ruleset
                )

        return InfoMessage(
            f"Connection to debugger established with context(\n\t{sample_hash},\n\t{utils.truncate_message(ruleset)}\n)"
        )

    def eval(self, expr: str) -> str | PopupMessage:
        log.info(f"[DEBUGGER] Evaluating {expr}")
        if not self.ctxs:
            return ErrorMessage("YARI is not ready to evaluate... Ignoring evaluation request")

        if self.context_rule_name:
            expr = f"{self.context_rule_name}|{expr}"

        result = ""
        for source, ctx in self.ctxs.items():
            res = ctx.eval(expr)
            log.debug(f"[DEBUGGER] source {source} returned value {res}")
            result += self.display_eval_response(source, res)
        return result

    @classmethod
    def display_eval_response(cls, source: str, value: Any) -> str:
        """Convert EvalResponse to string."""
        return f"- {cls.display_context_source(source)} -> {cls.display_py_object(value)}\n"

    @staticmethod
    def display_context_source(source: str) -> str:
        """Convert context source to string."""
        return f"context({source})"

    @staticmethod
    def display_py_object(value: Any) -> str:
        """Convert YrValue to string."""
        res = ""
        if isinstance(value, int):
            res = f"Integer ({value}, {hex(value)}) -> {bool(value)}"
        elif isinstance(value, str):
            res = f"String ({value}) -> {bool(value)}"
        elif isinstance(value, float):
            res = f"Float ({value}) -> {bool(value)}"
        else:
            log.warning(f"Unknown type of value {value}")

        return res

    @staticmethod
    async def get_samples_dir(ls: Any) -> Path | None:
        samples_dir_config = await utils.get_config_from_editor(ls, "yls.yari.samplesDirectory")
        log.debug(f"[DEBUGGER] Got {samples_dir_config=}")
        samples_dir_path = Path(samples_dir_config)

        if not samples_dir_path.is_dir():
            log.debug("[DEBUGGER] Samples dir does not exist or is not a directory")
            return None

        if len(list(samples_dir_path.iterdir())) == 0:
            log.debug("[DEBUGGER] Samples dir is empty")
            return None

        return samples_dir_path


class DebuggerProvider:
    """Singleton class providing Debugger object."""

    debugger = None

    @classmethod
    def instance(cls) -> Debugger:
        """Return singleton instance."""
        if cls.debugger is None:
            cls.debugger = Debugger()
        return cls.debugger



from __future__ import annotations

import logging
from typing import Any

import pygls.lsp.types as lsp_types
import yaramod
from pygls.workspace import Document

from yls import completion
from yls import utils
from yls.hookspecs import ErrorMessage
from yls.hookspecs import PopupMessage
from yls.plugin_manager_provider import PluginManagerProvider

log = logging.getLogger(__name__)


class Hoverer:
    def __init__(self, ls: Any) -> None:
        self.ls = ls
        self.selected_range: lsp_types.Range | None = None

    async def hover(self, params: lsp_types.TextDocumentPositionParams) -> lsp_types.Hover | None:

        document = self.ls.workspace.get_document(params.text_document.uri)
        position = params.position

        if self.ls.debug_hover:
            return utils.debug_hover(document, position)

        # Evaluate entire block of selected expressions
        if (
            self.selected_range
            and self.selected_range.start != self.selected_range.end
            and utils.position_in_range(position, self.selected_range)
        ):
            expr_range = utils.range_to_expression(document, self.selected_range)
            if not expr_range:
                return None

            expr = utils.text_from_range(document, expr_range)
            log.debug(f'[HOVER] Hover request with selected text "{expr}"')
            eval_result = await self.eval_expression(expr)
            log.debug(f"[HOVER] Evaluation returned {eval_result}")
            return lsp_types.Hover(
                contents=utils.markdown_content(
                    f"""{self.result_to_markdown(eval_result)}
*Expression*:
```
{utils.remove_whitespace(expr)}
```

"""
                )
            )

        token = utils.cursor_token(document, position)

        # Yaramod failed to parse but we can do textual fallback
        if not token:
            return lsp_types.Hover(
                contents=utils.markdown_content(self.get_cursor_documentation(document, position))
            )

        log.debug(f'[HOVER] Hover request with token "{token}" and type "{token.type}"')
        if token.type == yaramod.TokenType.StringId:
            return await self.hover_string(token, document, position)
        elif token.type == yaramod.TokenType.Id:
            yara_file = utils.yaramod_parse_file(document.path)
            if yara_file is None:
                return None

            for rule in yara_file.rules:
                if rule.name == token.text:
                    rule_doc = f"""*Rule name* = "{rule.name}"

```
{rule.text}
```

*Condition*:
```
{rule.condition.text}
```
"""

                    eval_result = await self.eval_expression(rule.name)
                    if eval_result:
                        rule_doc = f"{self.result_to_markdown(eval_result)}\n{rule_doc}"

                    return lsp_types.Hover(contents=utils.markdown_content(rule_doc))
        elif token.type in [
            yaramod.TokenType.ValueSymbol,
            yaramod.TokenType.StructureSymbol,
            yaramod.TokenType.FunctionSymbol,
            yaramod.TokenType.DictionarySymbol,
            yaramod.TokenType.ArraySymbol,
            yaramod.TokenType.Lt,
            yaramod.TokenType.Gt,
            yaramod.TokenType.Le,
            yaramod.TokenType.Ge,
            yaramod.TokenType.Eq,
            yaramod.TokenType.Neq,
            yaramod.TokenType.And,
            yaramod.TokenType.Or,
            yaramod.TokenType.Not,
        ]:
            return await self.hover_cursor(document, position)

        return None

    async def hover_cursor(
        self, document: Document, position: lsp_types.Position
    ) -> lsp_types.Hover | None:
        log.debug(f'[HOVER] Hover cursor "{document}" @ "{position}"')
        # Get the documentation for word under the cursor
        cursor_documentation = self.get_cursor_documentation(document, position)

        hover_string = ""

        yara_file = utils.yaramod_parse_file(document.path)
        if yara_file:
            expr = utils.cursor_expression(yara_file, position)
            if expr:
                eval_result = await self.eval_expression(expr.text)
                if eval_result:
                    hover_string += self.result_to_markdown(eval_result)

        hover_string += cursor_documentation

        return lsp_types.Hover(contents=utils.markdown_content(hover_string))

    def result_to_markdown(self, result: str) -> str:
        return f"*Evaluation result*:\n\n{result}\n********\n\n"

    async def eval_expression(self, expr: str) -> str:
        log.debug(f'[HOVER] Evaluate expression "{expr}"')

        # Evaluate the expression
        eval_result_collected = []
        eval_results = await utils.pluggy_results(
            PluginManagerProvider.instance().hook.yls_eval(ls=self.ls, expr=expr)
        )

        res_success = next((res for res in eval_results if not isinstance(res, ErrorMessage)), None)
        for i, eval_result in enumerate(eval_results):
            # Display errors
            if isinstance(eval_result, PopupMessage):
                # Show errors only if no debugger returned a valid result
                if not res_success:
                    if len(eval_results) > 1:
                        eval_result.message = f"{utils.DEBUGGER_SOURCES[i]}: {eval_result.message}"
                    eval_result.show(self.ls)

                log.debug(eval_result.message)
                continue

            if not eval_result:
                continue

            if len(eval_results) > 1:
                eval_result_collected.append(f"**{utils.DEBUGGER_SOURCES[i]}**:\n{eval_result}")
            else:
                eval_result_collected.append(eval_result)

        return "\n".join(eval_result_collected)

    def get_cursor_documentation(self, document: Document, position: lsp_types.Position) -> str:
        word = utils.cursor_symbol(document, position)
        log.debug(f'[HOVER] Hover request for symbol or function "{word}"')
        if word is None:
            return ""

        symbol = self.ls.completer.completion_cache.get_symbol(word)
        if symbol is None:
            return ""

        if isinstance(symbol, completion.Function):
            doc = symbol.hover_documentation()
        elif isinstance(symbol, completion.Value):
            doc = symbol.hover_string()
        else:
            return ""

        return doc

    async def hover_string(
        self, token: yaramod.Token, document: Document, position: lsp_types.Position
    ) -> lsp_types.Hover | None:
        """Generate Hover for string"""
        cur_rule = self.ls.get_current_rule(document.uri, position)
        if not cur_rule:
            return None

        # NOTE: This can be prob abstracted
        for string in cur_rule.strings:
            # NOTE: Check the type of string
            if string.identifier == token.pure_text:
                string_doc = f"{string.identifier} = {string.text}"

                eval_result = await self.eval_expression(string.identifier)
                if eval_result:
                    string_doc = f"{self.result_to_markdown(eval_result)}\n{string_doc}"

                return lsp_types.Hover(contents=utils.markdown_content(string_doc))

        return None



from __future__ import annotations

import logging
from typing import Any

import yaramod
from pluggy import PluginManager
from pygls.lsp.types import DocumentFormattingParams
from pygls.lsp.types import MessageType
from pygls.lsp.types import TextEdit
from pygls.workspace import Document

from yls import hookimpl
from yls import hookspecs
from yls import utils
from yls.debugger import DebuggerProvider
from yls.hookspecs import PluggyRes
from yls.hookspecs import PopupMessage

log = logging.getLogger(__name__)


def get_plugin_manager() -> PluginManager:
    """Construct new plugin manager instance."""
    log.info("Creating PluginManager")
    manager = PluginManager("yls")
    manager.add_hookspecs(hookspecs)
    manager.load_setuptools_entrypoints("yls")
    manager.register(YlsCorePlugin(), name="core")
    log.info(f"\t{manager.get_plugins()=}")
    log.info(f"\t{manager.list_name_plugin()=}")
    manager.trace.root.setwriter(print)
    return manager


class PluginManagerProvider:
    """Singleton class providing PluginManager parser object."""

    manager = None

    @classmethod
    def instance(cls) -> PluginManager:
        """Return singleton instance."""
        if cls.manager is None:
            cls.manager = get_plugin_manager()
        return cls.manager


class YlsCorePlugin:
    @hookimpl(trylast=True)
    def create_yaramod_instance(self) -> yaramod.Yaramod | None:
        return yaramod.Yaramod()

    @hookimpl(trylast=True)
    def yls_formatting(
        self,
        ls: Any,
        params: DocumentFormattingParams,  # pylint: disable=unused-argument
        document: Document,
    ) -> list[TextEdit]:
        # NOTE: Change after yaramod has support for "include" in parse_string
        yfile = utils.yaramod_parse_file(document.path)
        if yfile is None:
            ls.show_message(
                "Cannot parse the current document, please fix all errors before formatting.",
                MessageType.Error,
            )
            return []

        formatted_file = yfile.text_formatted

        total_lines = len(document.source.splitlines()) + 1
        res = [
            TextEdit(
                range=utils.range_from_coords((0, 0), (total_lines, 0)), new_text=formatted_file
            )
        ]
        return res

    @hookimpl
    def yls_scan_enabled(self) -> bool:
        return False

    @hookimpl(trylast=True)
    def yls_eval(
        self,  # pylint: disable=unused-argument
        ls: Any,  # pylint: disable=unused-argument
        expr: str,
    ) -> PluggyRes[str | PopupMessage]:
        return DebuggerProvider().instance().eval(expr)

    @hookimpl
    def yls_eval_set_context(
        self, ls: Any, _hash: str, ruleset: str  # pylint: disable=unused-argument
    ) -> PluggyRes[PopupMessage]:
        return DebuggerProvider().instance().set_context(ls, _hash, ruleset)

    @hookimpl
    def yls_eval_enabled(self) -> bool:
        return True



# pylint: disable=unused-argument

"""
Declaration of the available hooks for YLS.

WARNING: Declared API is unstable and is likely to change.
If you want to write a plugin for YLS feel free to contact us using issues or
other means.
"""

from __future__ import annotations

from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar
from typing import Union

import pluggy
import yaramod
from pygls.lsp.types import CodeAction
from pygls.lsp.types import CodeActionParams
from pygls.lsp.types import CodeLens
from pygls.lsp.types import CodeLensParams
from pygls.lsp.types import Command
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionParams
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DocumentFormattingParams
from pygls.lsp.types import MessageType
from pygls.lsp.types import TextEdit
from pygls.workspace import Document

hookspec = pluggy.HookspecMarker("yls")

T = TypeVar("T")
PluggyRes = Union[Awaitable[T], Callable[[], T], T]


@hookspec(firstresult=True)
def create_yaramod_instance() -> yaramod.Yaramod | None:
    ...


@hookspec
def yls_lint(document: Document) -> list[Diagnostic]:
    ...


@hookspec
def yls_completion(params: CompletionParams, document: Document) -> list[CompletionItem]:
    ...


@hookspec(firstresult=True)
def yls_formatting(
    ls: Any, params: DocumentFormattingParams, document: Document
) -> list[TextEdit] | None:
    ...


@hookspec
def yls_code_actions(ls: Any, params: CodeActionParams) -> list[Command | CodeAction]:
    ...


@hookspec
def yls_supported_commands(ls: Any) -> list[str]:
    ...


@hookspec
def yls_execute_command(command: str, ls: Any, params: Any) -> PluggyRes[Any | None]:
    ...


@hookspec
def yls_code_lens(ls: Any, params: CodeLensParams) -> list[CodeLens]:
    ...


@hookspec
def yls_scan(ls: Any, document: Document, rule_name: str | None) -> PluggyRes[list[Diagnostic]]:
    ...


@hookspec
def yls_scan_enabled() -> bool:
    ...


@hookspec
def yls_eval_enabled() -> bool:
    ...


class PopupMessage:
    def __init__(self, message: str, message_type: MessageType):
        self.message = message
        self.message_type = message_type

    def show(self, ls: Any) -> None:
        ls.show_message(self.message, self.message_type)


class InfoMessage(PopupMessage):
    def __init__(self, message: str):
        super().__init__(message, MessageType.Info)


class WarningMessage(PopupMessage):
    def __init__(self, message: str):
        super().__init__(message, MessageType.Warning)


class ErrorMessage(PopupMessage):
    def __init__(self, message: str):
        super().__init__(message, MessageType.Error)


@hookspec
def yls_eval(ls: Any, expr: str) -> PluggyRes[str | PopupMessage]:
    ...


@hookspec
def yls_eval_set_context(ls: Any, _hash: str, ruleset: str) -> PluggyRes[PopupMessage]:
    ...



from __future__ import annotations

import logging
import re

import yaramod
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DiagnosticSeverity
from pygls.workspace import Document

from yls import utils
from yls.plugin_manager_provider import PluginManagerProvider
from yls.yaramod_provider import YaramodProvider

log = logging.getLogger(__name__)


def get_diagnostics(document: Document) -> list[Diagnostic]:
    """Return diagnostic information for the given `Document`.

    This will perform various linting procedures and return a list of
    `Diagnostic`."""
    return Linter(document).lint()


class Linter:

    SOURCE_YARA = "Yara Python"
    SOURCE_YARAMOD = "Yaramod"

    SCHEMA_JSON_FILENAME = "schema.json"

    def __init__(self, document: Document):
        self.document = document

        # Result of the linting
        self.diag: list[Diagnostic] = []

    def lint(self) -> list[Diagnostic]:
        """Invoke linting procedure for each linter."""
        diagnostic = []

        diagnostic += utils.flatten_list(
            PluginManagerProvider.instance().hook.yls_lint(document=self.document)
        )
        diagnostic += self.lint_with_yaramod()

        return diagnostic

    def lint_with_yaramod(self) -> list[Diagnostic]:
        """Perform linting with Yaramod."""
        try:
            _ = YaramodProvider.instance().parse_file(self.document.path)
            log.debug("[LINT] Yaramod - no problems")
        except yaramod.ParserError as e:
            etext = str(e)
            log.debug(f'[LINT] Yaramod parse exception: "{etext}"')

            m = re.search(r"^Error at ([^ ]*):(\d+)\.(\d+)-?(\d*):(.*)", etext)
            if m:
                line = int(m.group(2)) - 1
                start = int(m.group(3)) - 1
                end = int(m.group(4)) if m.group(4) else start
                error_msg = m.group(5).strip()
                diagnostic = Diagnostic(
                    range=utils.range_from_coords((line, start), (line, end)),
                    message=error_msg,
                    source=self.SOURCE_YARAMOD,
                    severity=DiagnosticSeverity.Error,
                )
                return [diagnostic]

        return []
